import unittest
from bppy import *
from bppy.utils.dfs import DFSBThread, DFSBProgram
from bppy.utils.exceptions import BPAssertionError

class TestDFS(unittest.TestCase):

    def test_hello_world(self):
        @b_thread
        def hello():
            yield {request: BEvent("Hello")}

        dfs = DFSBThread(lambda: hello(), SimpleEventSelectionStrategy(), [BEvent("Hello"), BEvent("Dummy")])
        init_s, visited = dfs.run()
        assert len(visited) == 2

    def test_hot(self):
        @b_thread
        def add_hot():
            for i in range(3):
                yield {request: BEvent("HOT")}

        dfs = DFSBThread(lambda: add_hot(), SimpleEventSelectionStrategy(), [BEvent("HOT"), BEvent("COLD"), BEvent("Dummy")])
        init_s, visited = dfs.run()
        assert len(visited) == 4
    def test_hot_rb(self):
        @b_thread
        def add_hot():
            for i in range(3):
                yield {request: BEvent("HOT")}

        dfs = DFSBThread(lambda: add_hot(), SimpleEventSelectionStrategy(), [BEvent("HOT"), BEvent("COLD"), BEvent("Dummy")])
        init_s, visited, requested, blocked = dfs.run(return_requested_and_blocked=True)
        assert len(visited) == 4 and len(requested) == 1 and len(blocked) == 0

    def test_control(self):
        @b_thread
        def control_temp():
            e = BEvent("Dummy")
            while True:
                e = yield {waitFor: All(), block: e}

        dfs = DFSBThread(lambda: control_temp(), SimpleEventSelectionStrategy(), [BEvent("HOT"), BEvent("COLD")])
        init_s, visited = dfs.run()
        assert len(visited) == 3

    def test_bprogram(self):
        @b_thread
        def add_hot():
            for i in range(3):
                yield {request: BEvent("HOT")}

        @b_thread
        def add_cold():
            for i in range(3):
                yield {request: BEvent("COLD")}

        @b_thread
        def control():
            while True:
                yield {waitFor: BEvent("HOT")}
                yield {waitFor: All(), block: BEvent("HOT")}

        def bp_gen():
            return BProgram(bthreads=[add_hot(), add_cold(), control()],
                            event_selection_strategy=SimpleEventSelectionStrategy())

        dfs = DFSBProgram(bp_gen, [BEvent("HOT"), BEvent("COLD"), BEvent("Dummy")])
        init_s, visited = dfs.run()
        assert len(visited) == 19

    def test_bprogram_no_list(self):
        @b_thread
        def add_hot():
            for i in range(3):
                yield {request: BEvent("HOT")}

        @b_thread
        def add_cold():
            for i in range(3):
                yield {request: BEvent("COLD")}

        @b_thread
        def control():
            while True:
                yield {waitFor: BEvent("HOT")}
                yield {waitFor: All(), block: BEvent("HOT")}

        def bp_gen():
            return BProgram(bthreads=[add_hot(), add_cold(), control()],
                            event_selection_strategy=SimpleEventSelectionStrategy())

        dfs = DFSBProgram(bp_gen)
        init_s, visited = dfs.run()
        assert len(visited) == 19

    def test_bp_assertion_error(self):
        @b_thread
        def add_hot():
            for i in range(3):
                yield {request: BEvent("HOT")}

        @b_thread
        def add_cold():
            for i in range(3):
                yield {request: BEvent("COLD")}
            assert False

        @b_thread
        def control():
            while True:
                yield {waitFor: BEvent("HOT")}
                yield {waitFor: All(), block: BEvent("HOT")}

        def bp_gen():
            return BProgram(bthreads=[add_hot(), add_cold(), control()],
                            event_selection_strategy=SimpleEventSelectionStrategy())

        dfs = DFSBProgram(bp_gen)
        try:
            init_s, visited = dfs.run()
        except BPAssertionError:
            pass
        assert True




import unittest
from bppy import *
from bppy.gym import *
import numpy as np


class TestGym(unittest.TestCase):

    def test_env(self):
        @b_thread
        def add_hot():
            for i in range(3):
                yield {request: BEvent("HOT"), localReward: -0.1}
            yield {request: BEvent("DONE"), localReward: 1}

        @b_thread
        def add_cold():
            for i in range(3):
                yield {request: BEvent("COLD")}

        @b_thread
        def control():
            while True:
                yield {waitFor: BEvent("HOT")}
                yield {waitFor: BEvent("COLD"), block: BEvent("HOT")}

        def init_bprogram():
            return BProgram(bthreads=[add_hot(), add_cold(), control()],
                            event_selection_strategy=SimpleEventSelectionStrategy())

        env = BPEnv(bprogram_generator=init_bprogram,
                    event_list=[BEvent("HOT"), BEvent("COLD"), BEvent("DONE")])

        state, _ = env.reset()
        assert (state == np.asarray([0, 0, 0])).all()
        state, reward, terminated, _, info = env.step(0)
        assert (state == np.asarray([1, 0, 1])).all()



import unittest
from bppy import *
from bppy.analysis.dfs_bprogram_verifier import DFSBProgramVerifier
from bppy.analysis.symbolic_bprogram_verifier import SymbolicBProgramVerifier


class TestAnalysis(unittest.TestCase):

    def test_dfs_failed(self):
        @b_thread
        def add_hot():
            for i in range(3):
                yield {request: BEvent("HOT")}

        @b_thread
        def add_cold():
            for i in range(3):
                yield {request: BEvent("COLD")}

        @b_thread
        def control():
            while True:
                yield {waitFor: BEvent("HOT")}
                yield {waitFor: All(), block: BEvent("HOT")}

        @b_thread
        def check():
            yield {waitFor: BEvent("HOT")}
            yield {waitFor: BEvent("HOT")}
            yield {waitFor: BEvent("HOT")}
            assert False

        def bp_gen():
            return BProgram(bthreads=[add_hot(), add_cold(), control(), check()],
                            event_selection_strategy=SimpleEventSelectionStrategy())

        ver = DFSBProgramVerifier(bp_gen)
        ok, counter_example = ver.verify()
        assert not ok

    def test_dfs_passed(self):
        @b_thread
        def add_hot():
            for i in range(3):
                yield {request: BEvent("HOT")}

        @b_thread
        def add_cold():
            for i in range(3):
                yield {request: BEvent("COLD")}

        @b_thread
        def control():
            while True:
                yield {waitFor: BEvent("HOT")}
                yield {waitFor: All(), block: BEvent("HOT")}

        @b_thread
        def check():
            yield {waitFor: BEvent("HOT")}
            yield {waitFor: BEvent("HOT")}
            yield {waitFor: BEvent("HOT")}
            yield {waitFor: BEvent("HOT")}
            assert False

        def bp_gen():
            return BProgram(bthreads=[add_hot(), add_cold(), control(), check()],
                            event_selection_strategy=SimpleEventSelectionStrategy())

        ver = DFSBProgramVerifier(bp_gen)
        ok, counter_example = ver.verify()
        assert ok

    def test_symbolic_pass(self):

        @b_thread
        def add_hot():
            for i in range(3):
                yield {request: BEvent("HOT")}
            yield {request: BEvent("FINISH"), block: {BEvent("COLD")}}

        @b_thread
        def add_cold():
            for i in range(3):
                yield {request: BEvent("COLD")}

        @b_thread
        def control():
            while True:
                yield {waitFor: BEvent("HOT"), block: EventSet(lambda e: e.name == "COLD")}
                yield {waitFor: All(), block: BEvent("HOT")}

        def bp_gen():
            return BProgram(bthreads=[add_hot(), add_cold(), control()])

        verifier = SymbolicBProgramVerifier(bp_gen, [BEvent("HOT"), BEvent("COLD"), BEvent("FINISH")])
        result, explanation_str = verifier.verify(spec="F event = FINISH", type="BDD", bound=1000,
                                                  find_counterexample=True, print_info=False)

        assert result

    def test_symbolic_fail(self):

        @b_thread
        def add_hot():
            for i in range(3):
                yield {request: BEvent("HOT")}
            yield {request: BEvent("FINISH"), block: {BEvent("COLD")}}

        @b_thread
        def add_cold():
            for i in range(3):
                yield {request: BEvent("COLD")}

        @b_thread
        def control():
            while True:
                yield {waitFor: BEvent("HOT")}
                yield {waitFor: All(), block: BEvent("HOT")}

        def bp_gen():
            return BProgram(bthreads=[add_hot(), add_cold(), control()])

        verifier = SymbolicBProgramVerifier(bp_gen, [BEvent("HOT"), BEvent("COLD"), BEvent("FINISH")])
        result, explanation_str = verifier.verify(spec="F event = FINISH", type="BMC", bound=100,
                                                  find_counterexample=True, print_info=False)

        assert not result and explanation_str is not None

